'''Assetto Corsa Game Server Class'''

import asyncio
import configparser
import json
import logging
import os
import sys
from dataclasses import dataclass, field
from typing import Dict, List

from ac_websocket_server.child import ChildServer
from ac_websocket_server.entries import EntryList
from ac_websocket_server.error import WebsocketsServerError
from ac_websocket_server.grid import Grid
from ac_websocket_server.objects import (DriverInfo, EntryInfo, SessionInfo)
from ac_websocket_server.lobby import Lobby
from ac_websocket_server.protocol import Protocol
from ac_websocket_server.watcher import Watcher


@dataclass
class GameServer(ChildServer):
    '''Represents an Assetto Corsa Server.'''
    # pylint: disable=logging-fstring-interpolation, invalid-name

    directory: str
    child_ini_file: str = field(default='cfg/server_cfg.ini')
    child_title: str = field(default='AC Server')
    child_short: str = field(default='AC Server')
    is_optional: bool = False

    version: str = field(default='n/a', init=False)
    timestamp: str = field(default='n/a', init=False)
    name: str = field(init=False)
    track: str = field(init=False)
    cars: str = field(init=False)
    http_port: int = field(init=False)
    tcp_port: int = field(init=False)
    udp_port: int = field(init=False)
    drivers: Dict[str, DriverInfo] = field(init=False)
    entries: Dict[int, EntryInfo] = field(init=False)
    sessions: Dict[str, SessionInfo] = field(init=False)
    lobby: Lobby = field(init=False)
    running: int = field(init=False, default=False)

    def __post_init__(self):

        super().__post_init__()

        self._logger = logging.getLogger('ac-ws.game')

        self._args = ()

        if os.path.exists(f'{self.directory}/acServer.py'):
            self._cwd = None
            self._exe = f'{self.directory}/acServer.py'
            self._hash = None
        else:
            self._cwd = self.directory
            if sys.platform == 'linux':
                self._exe = f'{self.directory}/acServer'
                self._hash = 'f781ddfe02e68adfa170b28d0eccbbdc'
            else:
                self._exe = f'{self.directory}/acServer.exe'
                self._hash = '357e1f1fd8451eac2d567e154f5ee537'

        self.drivers = {}
        self.entries = {}
        self.sessions = {}

        self.parse_server_cfg()
        self.parse_entry_list()

        self.grid = Grid(server_directory=self.directory,
                         track=self.track,
                         entry_list=self.entry_list)

        self.lobby = Lobby()

        self._watcher_stdout: Watcher

    async def consumer(self, message_words: List[str], connection: id = None):
        '''Consume args destined for the server'''

        message_funcs = {'drivers': self.__drivers,
                         'entries': self.__entries,
                         'info': self.__info,
                         'sessions': self.__sessions,
                         'start': self.start,
                         'stop': self.stop,
                         'restart': self.restart}

        if message_funcs.get(message_words[0]):
            await message_funcs[message_words[0]]()

    async def __drivers(self):
        '''Show game drivers info as part of a JSON reply'''
        await self.put(Protocol.success({'drivers': self.drivers}))

    async def __entries(self):
        '''Show game entries info as part of a JSON reply'''
        await self.put(Protocol.success({'entries': self.entries}))

    async def __info(self):
        '''Show game server info as a JSON string'''
        await self.put(Protocol.success({'server': self}))

    async def notify(self, notifier):
        '''Receive a notification of a new message from log watcher or lobby.'''

        message = await notifier.get(self)

        try:
            item = json.loads(message)

            if not item.get('type', None):
                if item.get('data', None):
                    await self.put(Protocol.success(item['data']))
                if item.get('error', None):
                    await self.put(Protocol.error(item['error']))
                return

            if item['type'] == 'LobbyInfo':
                await self.lobby.update(item['body'])
                await self.put(Protocol.success({'lobby': item['body']}))

            if item['type'] == 'ServerInfo':
                self.version = item['body']['version']
                self.timestamp = item['body']['timestamp']
                self.track = item['body']['track']
                self.cars = item['body']['cars']

                await self.put(Protocol.success({'server': item['body']}))

            if item['type'] == 'DriverInfo' and item['body']['msg'] == 'joining':

                body = item['body']
                name = body['name']

                driver_info = DriverInfo()

                driver_info.name = name
                driver_info.host = body['host']
                driver_info.port = body['port']
                driver_info.car = body['car']
                driver_info.guid = body['guid']
                driver_info.ballast = body['ballast']
                driver_info.restrictor = body['restrictor']
                driver_info.msg = body['msg']

                self.drivers[driver_info.name] = driver_info

                self._logger.debug(f'Driver {name} joining')
                await self.put(Protocol.success({'driver': item['body']}))

            if item['type'] == 'DriverInfo' and item['body']['msg'] == 'leaving':
                body = item['body']
                name = body['name']
                del self.drivers[name]

                self._logger.debug(f'Driver {name} leaving')
                await self.put(Protocol.success({'driver': item['body']}))

            if item['type'] == 'SessionInfo':

                body = item['body']
                session_type = body['type']

                session_info = SessionInfo()

                session_info.type = session_type
                session_info.laps = body['laps']
                session_info.time = body['time']

                self.sessions[session_type] = session_info

                await self.put(Protocol.success({'sessions': session_info}))

        except json.JSONDecodeError:
            pass

        # await self.put(message)

    def parse_entry_list(self):
        '''Parse entry list file and update attributes'''

        entry_list_file_name = f'{self.directory}/cfg/entry_list.ini'

        if not os.path.exists(entry_list_file_name):
            error_message = f'Missing entry_list.ini file in {self.directory}'
            self._logger.error(error_message)
            raise WebsocketsServerError(error_message)

        self.entry_list = EntryList(entry_list_file_name)
        self.entries = self.entry_list.entries

    def parse_server_cfg(self):
        '''Parse server config file and update attributes'''

        server_cfg_file_name = f'{self.directory}/{self.child_ini_file}'

        if not os.path.exists(server_cfg_file_name):
            error_message = f'Missing server_cfg.ini file in {self.directory}'
            self._logger.error(error_message)
            raise WebsocketsServerError(error_message)

        cfg = configparser.ConfigParser()
        try:
            cfg.read(server_cfg_file_name)

            self.name = cfg['SERVER']['NAME']
            self.cars = cfg['SERVER']['CARS']
            self.track = cfg['SERVER']['TRACK']
            self.http_port = cfg['SERVER']['HTTP_PORT']
            self.tcp_port = cfg['SERVER']['TCP_PORT']
            self.udp_port = cfg['SERVER']['UDP_PORT']

            for session in ['PRACTICE', 'QUALIFY', 'RACE']:
                if cfg.has_section(session):
                    name = cfg[session].get('NAME')
                    time = cfg[session].get('TIME', 0)
                    laps = cfg[session].get('LAPS', 0)
                    self.sessions[name] = SessionInfo(
                        name, laps=laps, time=time)

        except configparser.Error as e:
            raise WebsocketsServerError(e) from e

    def pre_start_hook(self):
        '''Parse and update status before start (for re-start post __init__)'''

        super().pre_start_hook()

        try:
            self.parse_server_cfg()
            self.parse_entry_list()
            self.grid.track = self.track
        except WebsocketsServerError as e:
            self._logger.error('Command failed - configuration error')
            raise e

    async def post_start_hook(self):

        await self.put(Protocol.success(
            msg='Assetto Corsa server started'))

        self.lobby.subscribe(self)

        self._watcher_stdout = Watcher(self._logfile_stdout)
        self._watcher_stdout.subscribe(self)
        await self._watcher_stdout.start()

    async def post_stop_hook(self):

        await self.put(Protocol.success(
            msg='Assetto Corsa server stopped'))

        await self._watcher_stdout.stop()
        self._watcher_stdout.unsubscribe(self)

        self.lobby.unsubscribe(self)

    async def __sessions(self):
        '''Show game sessions info as a JSON string'''
        await self.put(Protocol.success({'sessions': self.sessions}))



#!/usr/bin/env python

'''Assetto Corsa Websockets App'''

import asyncio
from datetime import datetime
import json
import re
import sys
import tkinter as tk
from tkinter import ttk
from typing import Optional
import websockets

from ac_websocket_client import DEBUG
from ac_websocket_client.objects import *
from ac_websocket_server.protocol import Protocol


class App(tk.Tk):
    '''Wrapper class for Tk app'''

    UPDATE_INTERVAL = 1/120

    def __init__(self, loop, url=None):
        super().__init__()

        self.loop = loop
        self.protocol("WM_DELETE_WINDOW", self.stop_ui)

        self.consumer_queue = asyncio.Queue()
        self.producer_queue = asyncio.Queue()

        self.is_connected = False
        self.is_registered = False
        self.is_started = False
        self.is_tracking = False

        self.drivers = {}
        self.entries = {}

        self.url = url
        self.websocket = None

        self.acws_connect_button = tk.StringVar(value='Connect')
        self.ac_game_button = tk.StringVar(value='Start Game')
        self.ac_lobby_button = tk.StringVar(value='(Re)register')
        self.tracker_button = tk.StringVar(value='Start Tracker')
        self.toggle_grid_button = tk.StringVar(value='Show Active')

        self.acws_traffic_light: Optional[TrafficLight] = None
        self.ac_game_traffic_light: Optional[TrafficLight] = None
        self.ac_lobby_traffic_light: Optional[TrafficLight] = None
        self.tracker_traffic_light: Optional[TrafficLight] = None
        self.driver_saved_traffic_light: Optional[TrafficLight] = None
        self.console_sent_traffic_light: Optional[TrafficLight] = None

        self.url_entry = tk.StringVar()

        self.game_name_entry = tk.StringVar()
        self.game_timestamp_entry = tk.StringVar()
        self.registered_entry = tk.StringVar()
        self.track_entry = tk.StringVar()
        self.cars_entry = tk.StringVar()
        self.session_entry = tk.StringVar()

        self.tracker_timestamp_entry = tk.StringVar()

        self.console_command_button = tk.StringVar(value='Send')
        self.console_command_entry = tk.StringVar()

        if self.url:
            self.url_entry.set(self.url)

        self.tasks = []

        self._create_ui()

    def _create_ui(self):
        '''Build the UI elements'''

        self.title(APP_TITLE)
        self.config(bg='lightgray')
        self.columnconfigure(0, weight=1)

        # ACWS Frame
        self.rowconfigure(0, weight=1)
        acws_frame = GriddedFrame(grid_row=0, grid_col=0, height_by=0.5)
        acws_frame.configure_columns(1, 1, 4, 1, 1)

        GriddedLabel(acws_frame, grid_row=0, grid_col=0, width=8,
                     text="ACWS")
        GriddedLabel(acws_frame, grid_row=0, grid_col=1, width=8,
                     text="url:")
        connect_entry = GriddedEntry(acws_frame, grid_row=0, grid_col=2,
                                     textvariable=self.url_entry)
        connect_entry.bind('<Return>', self._toggle_connection_event)
        GriddedButton(acws_frame, grid_row=0, grid_col=3,
                      textvariable=self.acws_connect_button,
                      command=lambda: self.loop.create_task(self._toggle_connection()))
        self.acws_traffic_light = TrafficLight(acws_frame, row=0, column=4)

        # AC Frame
        self.rowconfigure(1, weight=4)
        ac_frame = GriddedFrame(grid_row=1, grid_col=0, height_by=2.5)
        ac_frame.configure_columns(1, 1, 4, 1, 1)

        GriddedLabel(ac_frame, grid_row=0, grid_col=0, width=8, text="Game")

        GriddedLabel(ac_frame, grid_row=0, grid_col=1,
                     width=8, text="started:")
        GriddedEntry(ac_frame, grid_row=0, grid_col=2,
                     textvariable=self.game_timestamp_entry, state=tk.DISABLED)

        GriddedLabel(ac_frame, grid_row=1, grid_col=1,
                     width=8, text="registered:")
        GriddedEntry(ac_frame, grid_row=1, grid_col=2,
                     textvariable=self.registered_entry, state=tk.DISABLED)

        GriddedLabel(ac_frame, grid_row=2, grid_col=1, width=8, text="name:")
        GriddedEntry(ac_frame, grid_row=2, grid_col=2,
                     textvariable=self.game_name_entry, state=tk.DISABLED)

        GriddedLabel(ac_frame, grid_row=3, grid_col=1, width=8, text="track:")
        GriddedEntry(ac_frame, grid_row=3, grid_col=2,
                     textvariable=self.track_entry, state=tk.DISABLED)

        GriddedLabel(ac_frame, grid_row=4, grid_col=1, width=8, text="cars:")
        GriddedEntry(ac_frame, grid_row=4, grid_col=2,
                     textvariable=self.cars_entry, state=tk.DISABLED)

        GriddedLabel(ac_frame, grid_row=5, grid_col=1,
                     width=8, text="session:")
        GriddedEntry(ac_frame, grid_row=5, grid_col=2,
                     textvariable=self.session_entry, state=tk.DISABLED)

        GriddedButton(ac_frame, grid_row=0, grid_col=3,
                      textvariable=self.ac_game_button,
                      command=lambda: self.loop.create_task(self._toggle_game()))
        self.ac_game_traffic_light = TrafficLight(ac_frame, row=0, column=4)

        GriddedButton(ac_frame, grid_row=1, grid_col=3,
                      textvariable=self.ac_lobby_button,
                      command=lambda: self.loop.create_task(self._toggle_registration()))
        self.ac_lobby_traffic_light = TrafficLight(ac_frame, row=1, column=4)

        # Tracker Frame
        self.rowconfigure(2, weight=1)
        tracker_frame = GriddedFrame(grid_row=2, grid_col=0, height_by=0.5)
        tracker_frame.configure_columns(1, 1, 4, 1, 1)

        GriddedLabel(tracker_frame, grid_row=0,
                     grid_col=0, width=8, text="Tracker")

        GriddedLabel(tracker_frame, grid_row=0,
                     grid_col=1, width=8, text="started:")
        GriddedEntry(tracker_frame, grid_row=0, grid_col=2,
                     textvariable=self.tracker_timestamp_entry, state=tk.DISABLED)

        GriddedButton(tracker_frame, grid_row=0, grid_col=3,
                      textvariable=self.tracker_button,
                      command=lambda: self.loop.create_task(self._toggle_tracker()))
        self.tracker_traffic_light = TrafficLight(
            tracker_frame, row=0, column=4)

        # Drivers Frame
        self.rowconfigure(3, weight=4)
        driver_frame = GriddedFrame(grid_row=3, grid_col=0, height_by=3)
        driver_frame.configure_columns(1, 1, 1, 1, 1, 1, 1)

        GriddedLabel(driver_frame, grid_row=0,
                     grid_col=0, width=8, text="Grid")

        GriddedButton(driver_frame, grid_row=0, grid_col=2,
                      textvar=self.toggle_grid_button, width=10,
                      command=lambda: self.loop.create_task(self._toggle_grid()))
        GriddedButton(driver_frame, grid_row=0, grid_col=3,
                      text='Order 1..n', width=10,
                      command=lambda: self.loop.create_task(self._set_grid(by_finishing=True)))
        GriddedButton(driver_frame, grid_row=0, grid_col=4,
                      text='Order n..1', width=10,
                      command=lambda: self.loop.create_task(self._set_grid(by_reverse=True)))
        GriddedButton(driver_frame, grid_row=0, grid_col=5,
                      text='Show', width=10,
                      command=lambda: self.loop.create_task(self._set_grid()))
        GriddedButton(driver_frame, grid_row=0, grid_col=6,
                      text='Save', width=10,
                      command=lambda: self.loop.create_task(self._set_grid(write=True)))
        self.driver_saved_traffic_light = TrafficLight(
            driver_frame, row=0, column=7)
        self.driver_saved_traffic_light.gray()

        self.driver_tree = GriddedTreeview(driver_frame, 1, 0, grid_span=8)
        self.driver_tree.add_columns('Name', 'GUID', 'Car', 'Ballast',
                                     'Restrictor', 'Position', 'Connected')
        self.driver_tree.set_widths(190, 80, 190, 80, 80, 80, 80)

        # Console Frame
        self.rowconfigure(4, weight=4)
        console_frame = GriddedFrame(grid_row=4, grid_col=0, height_by=3)
        console_frame.configure_columns(1, 1, 4, 1, 1)

        GriddedLabel(console_frame, grid_row=0, grid_col=0, width=8,
                     text="Console")
        GriddedLabel(console_frame, grid_row=0, grid_col=1, width=8,
                     text="command:")
        command_entry = GriddedEntry(console_frame, grid_row=0, grid_col=2,
                                     textvariable=self.console_command_entry)
        command_entry.bind('<Return>', self._send_command_event)

        GriddedButton(console_frame, grid_row=0, grid_col=3,
                      textvariable=self.console_command_button,
                      command=lambda: self.loop.create_task(self._send_command()))
        self.console_sent_traffic_light = TrafficLight(
            console_frame, row=0, column=4)
        self.console_sent_traffic_light.gray()

        listbox = GriddedListbox(
            console_frame, grid_row=1, grid_col=0, grid_span=5)

        self.tasks.append(self.loop.create_task(self._monitor(listbox)))

        if DEBUG:
            self.update()
            self._debug_ui(self, 0)

    def _debug_ui(self, w, depth=0):
        '''Print debug information on UI'''
        print('  ' * depth + f'{w.winfo_class()} w={str(w.winfo_width())}/{str(w.winfo_reqwidth())} h={str(w.winfo_height())}/{str(w.winfo_reqheight())} x/y=+{str(w.winfo_x())}+{str(w.winfo_y())}')
        for i in w.winfo_children():
            self._debug_ui(i, depth+1)

    async def _handler(self, websocket):
        '''Handle websocket tasks'''
        consumer_task = asyncio.create_task(self._handler_consumer(websocket))
        producer_task = asyncio.create_task(self._handler_producer(websocket))
        _done, pending = await asyncio.wait(
            [consumer_task, producer_task],
            return_when=asyncio.FIRST_COMPLETED,
        )
        for task in pending:
            task.cancel()

    async def _handler_consumer(self, websocket):
        '''Handle messages received from websocket'''
        async for msg in websocket:
            await self.consumer_queue.put(msg)

    async def _handler_producer(self, websocket):
        '''Handle messages received to send on websocket'''
        while True:
            try:
                message = await self.producer_queue.get()
                await websocket.send(message)
                await self.consumer_queue.put(Protocol.success(msg=f'Sent ==> {message}'))
            except Exception:
                print('\n> Connection Closing', file=sys.stderr)
                return

    async def _monitor(self, listbox: tk.Listbox):
        '''Monitor incoming messages and send to connection listbox'''
        while True:
            try:
                input_line = json.loads(await self.consumer_queue.get())

                output_fmt = {'fg': 'Black'}
                output_lines = ''

                if error_msg := input_line.get('error', None):

                    if error_msg['msg'] == 'ERROR,INVALID SERVER,CHECK YOUR PORT FORWARDING SETTINGS':
                        self.is_registered = False
                        self.ac_lobby_traffic_light.red()
                    output_fmt = {'fg': 'Red'}
                    output_lines = json.dumps(error_msg, indent=4)

                if success_msg := input_line.get('data', None):

                    if driver_msg := success_msg.get('driver', None):
                        if guid := driver_msg.get('guid', None):
                            if driver_joining := driver_msg.get('msg', None):
                                slot = driver_msg['slot']
                                self.entries[slot]['guid'] = guid
                                self.entries[slot]['drivername'] = driver_msg['name']
                                if 'joining' in driver_joining:
                                    self.drivers[guid] = driver_msg
                                    self.entries[slot]['connected'] = 'Yes'
                                    self._show_drivers()
                                else:
                                    self.entries[slot]['connected'] = 'No'
                                    del self.drivers[guid]
                                    self._show_drivers()

                    if session_msg := success_msg.get('lobby', None):
                        if session_msg['connected']:
                            self.is_registered = True
                            self.registered_entry.set(
                                datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
                            self.ac_lobby_traffic_light.green()
                        else:
                            self.is_registered = False
                            self.registered_entry.set('')
                            self.ac_lobby_traffic_light.red()

                    if server_msg := success_msg.get('server', None):
                        if name := server_msg.get('name', None):
                            self.game_name_entry.set(name)
                        if timestamp := server_msg.get('timestamp', None):
                            self.game_timestamp_entry.set(timestamp)
                        if server_msg.get('running', None):
                            self.is_started = True
                            self.ac_game_button.set('Stop Game')
                            self.ac_game_traffic_light.green()
                        self.track_entry.set(server_msg['track'])
                        self.cars_entry.set(server_msg['cars'])
                        if entries := server_msg.get('entries'):
                            for entry in entries:
                                slot = int(re.sub(r'\D', '', entry))
                                self.entries[slot] = entries[entry]
                                self.entries[slot]['connected'] = 'No'
                        self._show_entries()

                    if generic_msg := success_msg.get('msg', None):
                        if generic_msg == 'Stracker started':
                            self.tracker_traffic_light.green()
                        if generic_msg == 'Assetto Corsa server started':
                            self.ac_game_traffic_light.green()
                        if generic_msg == 'Assetto Corsa server stopped':
                            self.ac_game_traffic_light.red()

                    if session_msg := success_msg.get('sessions', None):
                        self.session_entry.set(
                            session_msg['type'] + ' ' + str(session_msg['laps']) + ' laps OR ' + str(session_msg['time']) + ' minutes')

                    output_fmt = {'fg': 'Green'}
                    output_lines = json.dumps(success_msg, indent=4)

                output_timestamp = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
                for output_line in output_lines.split('\n'):
                    listbox.insert(
                        tk.END, f'{output_timestamp}: {output_line}\n')
                    listbox.itemconfig(tk.END, output_fmt)
                listbox.yview(tk.END)

            except Exception as err:
                print(
                    f'Exception: "{err}" caught when processing:\n{input_line}')

    def _send_command_event(self, event):
        '''Send command to ACWS server'''
        self.loop.create_task(self._send_command())

    async def _send_command(self):
        '''Send command to ACWS server'''

        if self.is_connected:
            if command := self.console_command_entry.get():
                await self.consumer_queue.put(Protocol.success(msg=f'Sending ==> {command}'))
                await self.websocket.send(command)
        else:
            await self.consumer_queue.put(Protocol.error(msg='Not connected to ACWS server'))

    async def _set_grid(self, by_finishing: bool = None, by_reverse: bool = None, write: bool = None):
        '''Set the grid order'''

        if self.is_connected:
            if by_finishing:
                await self.websocket.send('grid finish')
                return
            if by_reverse:
                await self.websocket.send('grid reverse')
                return
            if write:
                await self.websocket.send('grid save')
                return
            await self.websocket.send('grid entries')
        else:
            await self.consumer_queue.put(Protocol.error(msg='Not connected to ACWS server'))

    def _show_drivers(self):
        # pylint: disable=consider-using-dict-items

        self.toggle_grid_button.set('Show Entries')
        self.driver_tree.delete(*self.driver_tree.get_children())
        for key in self.drivers:
            self.driver_tree.insert('', tk.END,
                                    values=(self.drivers[key]['name'],
                                            self.drivers[key]['guid'],
                                            self.drivers[key]['car'],
                                            self.drivers[key]['ballast'],
                                            self.drivers[key]['restrictor'],
                                            'n/a', 'Yes'))

    def _show_entries(self):
        # pylint: disable=consider-using-dict-items

        self.toggle_grid_button.set('Show Drivers')
        self.driver_tree.delete(*self.driver_tree.get_children())
        for key in self.entries:
            self.driver_tree.insert('', tk.END,
                                    values=(self.entries[key]['drivername'],
                                            self.entries[key]['guid'],
                                            self.entries[key]['model'],
                                            self.entries[key]['ballast'],
                                            self.entries[key]['restrictor'],
                                            str(key + 1),
                                            self.entries[key]['connected']))

    async def start_ui(self, interval=UPDATE_INTERVAL):
        '''Start a the update of the UI'''
        while True:
            self.update()
            await asyncio.sleep(interval)

    def stop_ui(self):
        '''Cleanup all tasks'''
        for task in self.tasks:
            task.cancel()
        self.loop.stop()
        self.destroy()

    def _toggle_connection_event(self, event):
        '''Connect to the websocket server'''
        self.loop.create_task(self._toggle_connection())

    async def _toggle_connection(self, url=None):
        '''Connect to the websocket server'''

        if not url:
            url = self.url_entry.get()

        if not url:
            url = self.url

        if self.is_connected:
            await self.websocket.close()
            return

        try:
            async with websockets.connect(url) as websocket:
                self.url = url
                self.url_entry.set(url)
                self.title(f'{APP_TITLE} - Connected to {url}')
                self.acws_connect_button.set('Disconnect')
                self.acws_traffic_light.green()
                self.is_connected = True
                self.websocket = websocket
                await self.producer_queue.put('server info')
                await self._handler(websocket)
                await self.consumer_queue.put(Protocol.success(msg=f'Disconnecting from {url}'))
                self.is_connected = False
                self.title(f'{APP_TITLE}')
                self.acws_connect_button.set('Connect')
                self.acws_traffic_light.red()
        except OSError as error:
            await self.consumer_queue.put(Protocol.error(msg=str(error)))

    async def _toggle_game(self):
        '''Start the game'''

        if self.is_connected:
            if not self.is_started:
                await self.websocket.send('server start')
                self.is_started = True
                self.ac_game_button.set('Stop Game')
            else:
                await self.websocket.send('server stop')
                self.is_started = False
                self.ac_game_button.set('Start Game')
                self.ac_game_traffic_light.red()
        else:
            await self.consumer_queue.put(Protocol.error(msg='Not connected to ACWS server'))

    async def _toggle_grid(self):

        if 'Show Drivers' in self.toggle_grid_button.get():
            self._show_drivers()
        else:
            self._show_entries()

    async def _toggle_registration(self):
        '''(Re)-register in lobby'''

        if self.is_connected:
            await self.websocket.send('lobby restart')
        else:
            await self.consumer_queue.put(Protocol.error(msg='Not connected to ACWS server'))

    async def _toggle_tracker(self):
        '''Toggle tracker'''

        if self.is_connected:
            if not self.is_tracking:
                await self.websocket.send('tracker start')
                self.is_tracking = True
                self.tracker_button.set('Stop Tracker')
            else:
                await self.websocket.send('tracker stop')
                self.is_tracking = False
                self.tracker_button.set('Start Tracker')
                self.tracker_traffic_light.red()
        else:
            await self.consumer_queue.put(Protocol.error(msg='Not connected to ACWS server'))



'''Assetto Corsa entry_list.ini helper Class to re-order and re-write'''

import configparser
import copy
import json
from enum import Enum, auto
import logging
import random
import sys
from typing import Any, Dict, List


from ac_websocket_server.objects import EntryInfo
from ac_websocket_server.error import WebsocketsServerError


class EntryListIterationMethod(Enum):
    '''Iteration method to rewrite file in various manners.'''
    ORIGINAL = auto()
    FINISHING = auto()
    REVERSE = auto()
    RANDOM = auto()


class EntryList:
    '''A collection of individual Entry for the entry_list.ini file'''

    def __init__(self, file_name: str = None, entries: Dict[int, EntryInfo] = None) -> None:
        '''
        Create a new EntryList with optional input file and pre-populated entries.
        '''

        self.__logger = logging.getLogger('ac-ws.entries')

        if entries:
            self.entries = entries
        else:
            self.entries = {}

        self.file_name = file_name

        if file_name and not entries:
            self.parse_entries_file()

        self.original_entries = copy.deepcopy(self.entries)

        self.entries_by_order: List[Any]
        self.index: int

        self.iteration_order = EntryListIterationMethod.ORIGINAL

    def __iter__(self):
        '''
        Create an Iterator based on EntryListIterationMethod
        '''

        self.entries_by_order = list(self.entries)
        self.index = 0

        if self.iteration_order == EntryListIterationMethod.ORIGINAL:
            self.entries_by_order.sort(
                key=lambda entry: self.entries[entry].old_id)

        if self.iteration_order == EntryListIterationMethod.FINISHING:
            self.entries_by_order.sort(
                key=lambda entry: self.entries[entry].new_id)

        if self.iteration_order == EntryListIterationMethod.REVERSE:
            self.entries_by_order.sort(
                key=lambda entry: self.entries[entry].new_id, reverse=True)

        if self.iteration_order == EntryListIterationMethod.RANDOM:
            for entry in self.entries.keys():
                self.entries[entry].new_id = random.randrange(100)
            self.entries_by_order.sort(
                key=lambda entry: self.entries[entry].new_id)

        return self

    def __next__(self):
        '''Return the next Entry'''

        if self.index < len(self.entries_by_order):
            entry = self.entries[self.entries_by_order[self.index]]
            self.index += 1
            return entry
        else:
            raise StopIteration

    def parse_entries_file(self):
        '''Parse the original entries file.'''

        if self.file_name:

            try:
                config = configparser.ConfigParser()
                config.read(self.file_name)

                for car_id in config.sections():

                    car = config[car_id]

                    self.entries[car_id] = \
                        EntryInfo(car_id=car_id,
                                  model=car.get('MODEL'),
                                  skin=car.get('SKIN'),
                                  spectator_mode=car.get('SPECTATOR_MODE'),
                                  drivername=car.get('DRIVERNAME'),
                                  team=car.get('TEAM'),
                                  guid=car.get('GUID'),
                                  ballast=car.get('BALLAST'),
                                  restrictor=car.get('RESTRICTOR', '')
                                  )

            except configparser.Error as error:
                raise WebsocketsServerError(error) from error

    def parse_result_file(self, result_file: str, track: str = None):
        '''Parse AC race results file'''
        # pylint: disable=invalid-name, logging-fstring-interpolation

        self.entries = copy.deepcopy(self.original_entries)

        try:
            with open(result_file, 'r', encoding='UTF-8') as f:
                data = json.load(f)
        except OSError as error:
            self.__logger.error(f'Unable to read input file: {result_file}')
            raise OSError from error

        if track and track != data["TrackName"]:
            raise WebsocketsServerError(
                f'TrackName from {result_file} does not match {track}')

        results = data["Result"]

        position = 1

        try:

            for result in results:

                car_id = 'CAR_' + str(result['CarId'])
                total_time = result['TotalTime']

                if total_time > 0:
                    self.entries[car_id].drivername = result['DriverName']
                    self.entries[car_id].guid = result['DriverGuid']
                    self.entries[car_id].new_id = position
                    position += 1
                else:
                    if self.iteration_order == EntryListIterationMethod.REVERSE:
                        self.entries[car_id].new_id = -sys.maxsize
                    else:
                        self.entries[car_id].new_id = sys.maxsize

        except KeyError as error:
            raise WebsocketsServerError(error) from error

    def __repr__(self):

        entries = ""

        position = 0
        for entry in iter(self):
            entry.new_id = position
            entries += str(entry)
            position += 1

        return entries

    def set_original_order(self):
        '''Set iteration to original cars order from entry list'''
        self.iteration_order = EntryListIterationMethod.ORIGINAL

    def set_standard_order(self):
        '''Set iteration to finishing order form last race'''
        self.iteration_order = EntryListIterationMethod.FINISHING

    def set_random_order(self):
        '''Set iteration to random'''
        self.iteration_order = EntryListIterationMethod.RANDOM

    def set_reversed_order(self):
        '''Set iteration to reversed grid from last race'''
        self.iteration_order = EntryListIterationMethod.REVERSE

    def show_entries(self) -> str:
        '''Returns a dictionary representation of all cars from entry_list.ini'''

        original_iteration_method = self.iteration_order

        self.iteration_order = EntryListIterationMethod.ORIGINAL

        entries = {}

        car_number = 1
        for entry in iter(self):
            if entry.drivername:
                driver_details = f'{entry.drivername} ({entry.guid})'
            else:
                driver_details = f'({entry.guid})'
            entries[str(car_number)] = f'{entry.model} {driver_details}'
            car_number += 1

        self.iteration_order = original_iteration_method

        return entries

    def show_grid(self) -> Dict[str, Any]:
        '''Returns a dictionary representation of the grid by player'''

        entries = {}

        position = 1
        for entry in iter(self):
            if entry.drivername != '':
                entries[str(position)] = entry.drivername
                position += 1

        return entries

    def write(self, output_file):
        '''Write updated entry list to output_file'''
        # pylint: disable=invalid-name, logging-fstring-interpolation

        try:
            with open(output_file, 'w', encoding='utf-8') as f:
                f.write(str(self))
                f.close()
        except OSError as err:
            self.__logger.error(f'Unable to write output file: {output_file}')
            raise OSError from err



'''Assetto Corsa Websocket Protocol Class'''

import json

from ac_websocket_server.objects import EnhancedJSONEncoder


class Protocol:
    '''Google JSON style protocol factory'''

    @classmethod
    def success(cls, data: any = None, msg: str = None) -> str:
        '''Return a success response with data or data message'''
        if msg:
            return json.dumps({'data': {'msg': msg}},
                              cls=EnhancedJSONEncoder, indent=4)
        return json.dumps({'data': data},
                          cls=EnhancedJSONEncoder, indent=4)

    @classmethod
    def error(cls, error: any = None, msg: str = None) -> str:
        '''Return an error response with error or error message'''
        if msg:
            return json.dumps({'error': {'msg': msg}},
                              cls=EnhancedJSONEncoder, indent=4)
        return json.dumps({'error': error},
                          cls=EnhancedJSONEncoder, indent=4)



'''Assetto Corsa Websocket Server'''

import logging
import sys

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s %(levelname)s %(message)s')

logger = logging.getLogger('ac-ws')

logging.getLogger("asyncio").setLevel(logging.DEBUG)
